from dropland.engines.redis import USE_REDIS

if USE_REDIS:
    from dropland.engines.redis.containers import RedisContainer
    from dropland.engines.redis.engine import EngineConfig

from tests import REDIS_URI

if USE_REDIS:
    redis_container = RedisContainer()
    redis_engine = redis_container.create_engine('dropland', EngineConfig(url=REDIS_URI))



from dependency_injector import containers, providers

from dropland.app.base import ContainerBlock, SessionResourceBlock
from dropland.util import default_value
from .engine import EngineConfig, SqlEngineBackend


class SqlBlock(ContainerBlock, SessionResourceBlock):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sync_initialized = self._async_initialized = False
        self._sync_sessions = list()
        self._async_sessions = list()

    def sync_startup(self, *args, **kwargs):
        if self._sync_initialized:
            return

        for engine in self.container.engine_factory().get_engines().values():
            if not engine.is_async:
                engine.start()

        self._sync_initialized = True

    def sync_shutdown(self, *args, **kwargs):
        if not self._sync_initialized:
            return

        for engine in reversed(self.container.engine_factory().get_engines().values()):
            if not engine.is_async:
                engine.stop()

        self._sync_initialized = False

    async def startup(self, *args, **kwargs):
        if self._async_initialized:
            return

        for engine in self.container.engine_factory().get_engines().values():
            if engine.is_async:
                await engine.start()

        self._async_initialized = True

    async def shutdown(self, *args, **kwargs):
        if not self._async_initialized:
            return

        for engine in reversed(self.container.engine_factory().get_engines().values()):
            if engine.is_async:
                await engine.stop()

        self._async_initialized = False

    def sync_session_begin(self, begin_tx: bool = True, autocommit: bool = True, *args, **kwargs):
        if not self._sync_initialized:
            return

        for engine in self.container.engine_factory().get_engines().values():
            if not engine.is_async:
                session = engine.session(begin_tx=begin_tx, autocommit=autocommit)
                session.__aenter__()
                self._sync_sessions.append(session)

    def sync_session_finish(self, *args, **kwargs):
        if not self._sync_initialized:
            return

        for session in reversed(self._sync_sessions):
            session.__exit__(None, None, None)
        self._sync_sessions.clear()

        for engine in reversed(self.container.engine_factory().get_engines().values()):
            if not engine.is_async:
                engine.teardown_session()

    async def session_begin(self, begin_tx: bool = True, autocommit: bool = True, *args, **kwargs):
        if not self._async_initialized:
            return

        for engine in self.container.engine_factory().get_engines().values():
            if engine.is_async:
                session = engine.session(begin_tx=begin_tx, autocommit=autocommit)
                await session.__aenter__()
                self._async_sessions.append(session)

    async def session_finish(self, *args, **kwargs):
        if not self._async_initialized:
            return

        for session in reversed(self._async_sessions):
            await session.__aexit__(None, None, None)
        self._async_sessions.clear()

        for engine in reversed(self.container.engine_factory().get_engines().values()):
            if engine.is_async:
                await engine.teardown_session()

    def get_engine(self, *args, **kwargs):
        if isinstance(self.container, SingleSqlContainer) or 'SingleSqlContainer' == self.container.parent_name:
            return self.container.create_engine()
        return self.container.create_engine(*args, **kwargs)


class SqlContainer(containers.DeclarativeContainer):
    __self__ = providers.Self()
    engine_factory = providers.Singleton(SqlEngineBackend)

    def _create_engine(self, *args, **kwargs):
        return self.engine_factory().create_engine(*args, **kwargs)

    create_engine = providers.Factory(_create_engine, __self__)
    block = providers.Singleton(SqlBlock, __self__)


class SingleSqlContainer(SqlContainer):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self):
        if isinstance(self.config.engine_config(), EngineConfig):
            engine_config = self.config.engine_config()
        else:
            engine_config = EngineConfig(
                url=self.config.engine_config.url(),
                echo=self.config.engine_config.echo.as_(bool)(),
                pool_min_size=self.config.engine_config.
                    pool_min_size.as_(default_value(int))(default=1),
                pool_max_size=self.config.engine_config.
                    pool_max_size.as_(default_value(int))(default=8),
                pool_expire_seconds=self.config.engine_config.
                    pool_expire_seconds.as_(default_value(int))(default=60),
                pool_timeout_seconds=self.config.engine_config.
                    pool_timeout_seconds.as_(default_value(int))(default=15)
            )
        return SqlContainer._create_engine(
            self, self.config.name(), engine_config, self.config.db_type(),
            use_async=self.config.use_async.as_(bool)())

    create_engine = providers.Factory(_create_engine, __self__)
    block = providers.Singleton(SqlBlock, __self__)


class MultipleSqlContainer(SqlContainer):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self, name: str):
        if conf := self.config.get(name):
            if isinstance(conf['engine_config'], EngineConfig):
                engine_config = conf['engine_config']
            else:
                engine_config = EngineConfig(
                    url=conf['engine_config']['url'],
                    echo=bool(conf['engine_config'].get('echo')),
                    pool_min_size=int(conf['engine_config'].get('pool_min_size', 1)),
                    pool_max_size=int(conf['engine_config'].get('pool_max_size', 8)),
                    pool_expire_seconds=int(conf['engine_config'].get('pool_expire_seconds', 60)),
                    pool_timeout_seconds=int(conf['engine_config'].get('pool_timeout_seconds', 15))
                )
            return SqlContainer._create_engine(
                self, name, engine_config, conf['db_type'], use_async=conf.get('use_async', False))
        return None

    create_engine = providers.Factory(_create_engine, __self__)
    block = providers.Singleton(SqlBlock, __self__)



import inspect


def _skip_redundant(iterable, skipset=None):
    """Redundant items are repeated items or items in the original skipset."""
    if skipset is None:
        skipset = set()
    for item in iterable:
        if item not in skipset:
            skipset.add(item)
            yield item


def _remove_redundant(metaclasses):
    skipset = {type}
    for meta in metaclasses:  # determines the metaclasses to be skipped
        skipset.update(inspect.getmro(meta)[1:])
    return tuple(_skip_redundant(metaclasses, skipset))


_memoized_metaclasses_map = {}


def _get_noconflict_metaclass(bases, left_metas, right_metas):
    """Not intended to be used outside of this module, unless you know
    what you are doing."""
    # make tuple of needed metaclasses in specified priority order
    metas = left_metas + tuple(map(type, bases)) + right_metas
    needed_metas = _remove_redundant(metas)

    # return existing confict-solving meta, if any
    if needed_metas in _memoized_metaclasses_map:
        return _memoized_metaclasses_map[needed_metas]

    # nope: compute, memoize and return needed conflict-solving meta
    elif not needed_metas:  # wee, a trivial case, happy us
        meta = type

    # another trivial case
    elif len(needed_metas) == 1:
        meta = needed_metas[0]

    # check for recursion, can happen i.e. for Zope ExtensionClasses
    elif needed_metas == bases:
        raise TypeError('Incompatible root metatypes', needed_metas)

    else:  # gotta work ...
        metaname = '_' + ''.join([m.__name__ for m in needed_metas])
        meta = classmaker()(metaname, needed_metas, {})

    _memoized_metaclasses_map[needed_metas] = meta
    return meta


def classmaker(left_metas=(), right_metas=()):
    def make_class(name, bases, adict):
        metaclass = _get_noconflict_metaclass(bases, left_metas, right_metas)
        return metaclass(name, bases, adict)
    return make_class



from abc import ABC, abstractmethod
from contextlib import contextmanager, asynccontextmanager, AbstractContextManager, AbstractAsyncContextManager
from typing import Optional, Union, List, Mapping, Callable, TypeVar, Generic

SessionType = TypeVar('SessionType')


class EngineBackend(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @abstractmethod
    def create_engine(self, name: str, *args, **kwargs) -> Union['SyncEngine', 'AsyncEngine']:
        ...

    @abstractmethod
    def get_engine(self, name: str) -> Optional[Union['SyncEngine', 'AsyncEngine']]:
        ...

    @abstractmethod
    def get_engine_names(self) -> List[str]:
        ...

    @abstractmethod
    def get_engines(self, names: Optional[List[str]] = None) -> Mapping[str, Union['SyncEngine', 'AsyncEngine']]:
        ...


class SyncEngine(Generic[SessionType], ABC):
    is_sync = True
    is_async = False

    def __init__(self, backend: EngineBackend):
        self._backend = backend

    @property
    def backend(self):
        return self._backend

    @contextmanager
    @abstractmethod
    def session(self, *args, **kwargs) -> Callable[..., AbstractContextManager[SessionType]]:
        ...

    def teardown_session(self):
        pass

    @abstractmethod
    def start(self):
        ...

    @abstractmethod
    def stop(self):
        ...


class AsyncEngine(Generic[SessionType], ABC):
    is_sync = False
    is_async = True

    def __init__(self, backend: EngineBackend):
        self._backend = backend

    @property
    def backend(self):
        return self._backend

    @asynccontextmanager
    @abstractmethod
    async def session(self, *args, **kwargs) -> Callable[..., AbstractAsyncContextManager[SessionType]]:
        ...

    async def teardown_session(self):
        pass

    @abstractmethod
    async def start(self):
        ...

    @abstractmethod
    async def stop(self):
        ...
